from abc import ABCMeta, abstractmethod
from collections.abc import MutableMapping, Sequence
from typing import Any, ClassVar, override

from .attrs import GlobalAttrs
from .base import BaseElement
from .elements import Blank as Blank
from .elements import Element, ElementStrict
from .html import div, span
from .styles import Theme, get_default_theme
from .styles.types import GlobalStyles
from .types import AnyChildren, TAttrs, TChildren, TChildrenArgs
from .utils import get_element_attrs_annotations

COMPONENT_REGISTRY: MutableMapping[str, list[type["BaseComponent"]]] = {}


class BaseComponent(BaseElement, metaclass=ABCMeta):
    classes: ClassVar[Sequence[str]] = []
    styles: ClassVar[GlobalStyles] = {}

    @property
    def theme(self) -> Theme:
        """Get the theme of the element."""
        if context_theme := self.context.get("theme"):
            if isinstance(context_theme, Theme):
                return context_theme
        return get_default_theme()

    def __init_subclass__(cls) -> None:
        COMPONENT_REGISTRY.setdefault(cls.__name__, [])
        COMPONENT_REGISTRY[cls.__name__].append(cls)

    def _add_classes(self, classes: list[str], element: BaseElement) -> None:
        if classes:
            element.attrs.setdefault("classes", [])  # type: ignore
            element.attrs["classes"].extend(classes)

    def attrs_for(self, cls: type["BaseElement"]) -> dict[str, Any]:
        """Get the attributes of this component that are defined in the given element.

        This is useful so that you can pass common attributes to an element
        without having to pass them from a parent one by one.

        Args:
            cls (type[BaseElement]): The element to get the attributes of.

        """
        return {
            key: value
            for key, value in self.attrs.items()
            if key in get_element_attrs_annotations(cls)
        }

    def to_html(self) -> str:
        dom: BaseElement | BaseComponent = self
        classes: list[str] = []

        while isinstance(dom, BaseComponent):
            classes += dom.classes
            context = dom.context
            dom = dom.render()
            dom.context.update(context)

        self._add_classes(classes, dom)
        return dom.to_html()

    @abstractmethod
    def render(self) -> BaseElement:
        """Render the component as an instance of :class:`BaseElement`."""


class Component(Element[TChildren, TAttrs], BaseComponent):
    """Base class for components.

    A component subclasses an :class:`Element` and represents any element
    that can be rendered in Ludic.

    Example usage:

        class PersonAttrs(Attributes):
            age: NotRequired[int]

        class Person(Component[PersonAttrs]):
            @override
            def render(self) -> dl:
                return dl(
                    dt("Name"),
                    dd(self.attrs["name"]),
                    dt("Age"),
                    dd(self.attrs.get("age", "N/A")),
                )

    Now the component can be used in any other component or element:

        >>> div(Person(name="John Doe", age=30), id="person-detail")

    """


class ComponentStrict(ElementStrict[*TChildrenArgs, TAttrs], BaseComponent):
    """Base class for strict components.

    A component subclasses an :class:`ElementStrict` and represents any
    element that can be rendered in Ludic. The difference between
    :class:`Component` and :class:`ComponentStrict` is that the latter
    expects concrete types of children. It allows specification
    of each child's type.

    Example usage:

        class PersonAttrs(Attributes):
            age: NotRequired[int]

        class Person(ComponentStrict[str, str, PersonAttrs]):
            @override
            def render(self) -> dl:
                return dl(
                    dt("Name"),
                    dd(" ".join(self.children),
                    dt("Age"),
                    dd(self.attrs.get("age", "N/A")),
                )

    Valid usage would look like this:

        >>> div(Person("John", "Doe", age=30), id="person-detail")

    In this case, we specified that Person expects two string children.
    First child is the first name, and the second is the second name.
    We also specify age as an optional key-word argument.
    """


class Block(Component[AnyChildren, GlobalAttrs]):
    """Component rendering as a div."""

    @override
    def render(self) -> div:
        return div(*self.children, **self.attrs)


class Inline(Component[AnyChildren, GlobalAttrs]):
    """Component rendering as a span."""

    @override
    def render(self) -> span:
        return span(*self.children, **self.attrs)



from typing import Annotated, Literal, TypedDict

import pytest
from starlette.datastructures import FormData

from ludic.catalog.forms import FieldMeta
from ludic.web.parsers import ListParser, Parser, ValidationError


def parse_bool(value: Literal["on", "off"]) -> bool:
    return True if value == "on" else False


class Example(TypedDict):
    sample_str: Annotated[str, FieldMeta()]
    sample_int: Annotated[int, FieldMeta(parser=int)]
    sample_bool: Annotated[bool, FieldMeta(parser=parse_bool)]


class ExampleOptional(TypedDict, total=False):
    sample_optional: Annotated[str, FieldMeta()]


class InvalidExample(TypedDict):
    sample_invalid: Annotated[int, FieldMeta()]  # missing parser=int


def test_parse_form_data() -> None:
    data = FormData({"sample_str": "test", "sample_int": "10", "sample_bool": "on"})
    assert Parser[Example](data).validate() == Example(
        sample_str="test",
        sample_int=10,
        sample_bool=True,
    )


def test_parse_invalid_form_data() -> None:
    data = FormData({"sample_str": "test", "sample_int": "10a", "sample_bool": "on"})
    with pytest.raises(ValidationError):
        _ = Parser[Example](data).validate()


def test_parse_invalid_example() -> None:
    data = FormData({"sample_invalid": "10"})
    with pytest.raises(ValidationError):
        _ = Parser[InvalidExample](data).validate()


def test_parse_empty_valid() -> None:
    data = FormData({})
    assert Parser[ExampleOptional](data).validate() == {}


def test_parse_list_form_data() -> None:
    data = FormData(
        {
            "sample_str:_index:0": "test2",
            "sample_int:_index:0": "90",
            "sample_bool:_index:0": "off",
            "sample_str:_index:1": "test1",
            "sample_int:_index:1": "10",
            "sample_bool:_index:1": "on",
        }
    )
    assert ListParser[Example](data).validate() == [
        Example(
            sample_str="test2",
            sample_int=90,
            sample_bool=False,
        ),
        Example(
            sample_str="test1",
            sample_int=10,
            sample_bool=True,
        ),
    ]


def test_parse_list_invalid_form_data() -> None:
    data = FormData(
        {
            "sample_str:_index:0": "test2",
            "sample_int:_index:0": "90abc",
            "sample_bool:_index:0": "off",
        }
    )
    with pytest.raises(ValidationError):
        _ = ListParser[Example](data).validate()


def test_parse_list_invalid_example() -> None:
    data = FormData({"sample_invalid:_intex:0": "10"})
    with pytest.raises(ValidationError):
        _ = ListParser[InvalidExample](data).validate()


def test_parse_list_empty_invalid() -> None:
    data = FormData({"sample_str": "10"})
    with pytest.raises(ValidationError):
        _ = ListParser[Example](data).validate()


def test_parse_list_empty_valid() -> None:
    data = FormData({})
    assert ListParser[ExampleOptional](data).validate() == []



from typing import override

from ludic.attrs import GlobalAttrs, ImgAttrs
from ludic.components import Component, ComponentStrict
from ludic.html import img, span, style
from ludic.types import AnyChildren, NoChildren


class Icon(Component[NoChildren, ImgAttrs]):
    classes = ["icon"]
    styles = {
        ".icon": {
            "width": "0.75em",
            "height": "0.75em",
        },
    }

    @override
    def render(self) -> img:
        return img(*self.children, **self.attrs)


class WithIcon(ComponentStrict[Icon, AnyChildren, GlobalAttrs]):
    classes = ["with-icon"]
    styles = style.use(
        lambda theme: {
            ".with-icon": {
                "display": "inline-flex",
                "align-items": "baseline",
            },
            ".with-icon .icon": {
                "margin-inline-end": theme.sizes.s,
            },
        }
    )

    @override
    def render(self) -> span:
        return span(*self.children, **self.attrs)



import functools
import inspect
from collections.abc import Callable
from typing import Any, ParamSpec

from fastapi._compat import lenient_issubclass
from fastapi.datastructures import Default, DefaultPlaceholder
from fastapi.dependencies.utils import get_typed_return_annotation
from fastapi.routing import APIRoute
from starlette._utils import is_async_callable
from starlette.responses import Response
from starlette.routing import get_name

from ludic.base import BaseElement
from ludic.web.responses import LudicResponse, run_in_threadpool_safe

P = ParamSpec("P")


def function_wrapper(
    handler: Callable[..., Any], status_code: int = 200
) -> Callable[P, Any]:
    """Wraps endpoints to ensure responses are formatted as LudicResponse objects.

    This function determines whether the handler is asynchronous or synchronous and
    executes it accordingly. If the handler returns a BaseElement instance, it wraps
    the response in a LudicResponse object.

    Args:
        handler (Callable[..., Any]): The FastAPI endpoint handler function.
        status_code (int, optional): The HTTP status code for the response.
            Defaults to 200.

    Returns:
        Callable[P, Any]: A wrapped function that ensures proper response handling.
    """

    @functools.wraps(handler)
    async def wrapped_endpoint(*args: P.args, **kwargs: P.kwargs) -> Any:
        if is_async_callable(handler):
            with BaseElement.formatter:
                raw_response = await handler(*args, **kwargs)
        else:
            raw_response = await run_in_threadpool_safe(handler, *args, **kwargs)

        if isinstance(raw_response, BaseElement):
            return LudicResponse(raw_response, status_code=status_code)
        return raw_response

    return wrapped_endpoint


class LudicRoute(APIRoute):
    """Custom Route class for FastAPI that integrates Ludic framework response handling.

    This class ensures that endpoints returning `BaseElement` instances are properly
    wrapped in `LudicResponse`. If a response model is not explicitly provided, it
    infers the return type annotation from the endpoint function.

    Args:
        path (str): The API route path.
        endpoint (Callable[..., Any]): The FastAPI endpoint function.
        response_model (Any, optional): The response model for OpenAPI documentation.
            Defaults to None.
        name (str | None, optional): The route name. Defaults to None.
        status_code (int | None, optional): The HTTP status code. Defaults to None.
        **kwargs (Any): Additional parameters for APIRoute.
    """

    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        response_model: Any = Default(None),  # noqa
        name: str | None = None,
        status_code: int | None = None,
        **kwargs: Any,
    ) -> None:
        if isinstance(response_model, DefaultPlaceholder):
            return_annotation = get_typed_return_annotation(endpoint)
            if lenient_issubclass(return_annotation, BaseElement) or lenient_issubclass(
                return_annotation, Response
            ):
                response_model = None
            else:
                response_model = return_annotation

        name = get_name(endpoint) if name is None else name
        wrapped_route = endpoint
        if inspect.isfunction(endpoint) or inspect.ismethod(endpoint):
            wrapped_route = function_wrapper(endpoint, status_code=status_code or 200)

        super().__init__(
            path, wrapped_route, response_model=response_model, name=name, **kwargs
        )



from fastapi import FastAPI, Response
from fastapi.testclient import TestClient

from ludic.base import BaseElement
from ludic.contrib.fastapi import LudicRoute
from ludic.html import div, p, span

app = FastAPI()
app.router.route_class = LudicRoute


@app.get("/component")
async def return_component() -> div:
    return div(span("Hello"), p("World"))


@app.get("/dict", response_model=None)
async def return_dict() -> dict[str, str]:
    return {"message": "Hello World"}


@app.get("/custom-response")
async def return_custom_response() -> Response:
    return Response(content="Custom", media_type="text/plain")


@app.get("/nested")
async def nested_components() -> div:
    return div(p("Level 1", span("Level 2", div("Level 3"))), id="root")


@app.get("/f-string")
async def f_string_test() -> span:
    return span(f"Result: {p('This is a test')}")


@app.get("/status-code", status_code=202)
async def status_code_test() -> span:
    return span("this is a test")


def test_auto_component_conversion() -> None:
    client = TestClient(app)
    response = client.get("/component")

    assert response.status_code == 200
    assert response.content == b"<div><span>Hello</span><p>World</p></div>"
    assert response.headers["content-type"] == "text/html; charset=utf-8"


def test_non_component_responses() -> None:
    client = TestClient(app)

    response = client.get("/dict")
    assert response.status_code == 200
    assert response.json() == {"message": "Hello World"}

    response = client.get("/custom-response")
    assert response.status_code == 200
    assert response.content == b"Custom"
    assert response.headers["content-type"] == "text/plain; charset=utf-8"


def test_nested_components() -> None:
    client = TestClient(app)
    response = client.get("/nested")

    assert response.status_code == 200
    assert (
        b'<div id="root"><p>Level 1<span>Level 2<div>Level 3</div></span></p></div>'
        in response.content
    )


def test_formatter_cleanup_with_f_string() -> None:
    client = TestClient(app)
    response = client.get("/f-string")

    assert response.status_code == 200
    assert b"Result: <p>This is a test</p>" in response.content
    assert len(BaseElement.formatter.get()) == 0


def test_return_different_status_code() -> None:
    client = TestClient(app)
    response = client.get("/status-code")

    assert response.status_code == 202
